/*
 * Export JZ4740 GPIO pins as a SPI server.
 *
 * Copyright (C) 2018, 2020, 2023, 2024 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/gpio-jz4740.h>
#include <l4/devices/spi-gpio.h>
#include <l4/devices/memory.h>

#include <l4/re/env.h>

#include <errno.h>
#include <stdint.h>
#include <stdlib.h>

#include <ipc/server.h>

#include "spi_server.h"



/* Virtual addresses for the GPIO register block. */

static l4_addr_t gpio_virt_base = 0, gpio_virt_base_end = 0;



static int setup_memory(void)
{
  if (get_memory("jz4740-gpio", &gpio_virt_base, &gpio_virt_base_end))
    return 1;

  return 0;
}



/* SPI peripheral device. */

class SPI_server : public SPI
{
  Spi_gpio *_spi;

public:
  /* Associate the device with a particular memory region. */

  explicit SPI_server(Spi_gpio *spi)
  : _spi(spi)
  {
  }

  /* Send a SPI command. */

  long send(int bits, int data)
  {
    uint32_t bytes = (bits + 7) / 8;
    uint8_t buffer[bytes];

    /* Convert the data into a sequence of bytes. */

    for (int byte = bytes; byte > 0; byte--)
      buffer[byte - 1] = (data >> ((byte - 1) * 8)) & 0xff;

    _spi->send(bytes, buffer);
    return L4_EOK;
  }
};



/*
Parse a string of the form "<port><pin>" where <port> is a single character
whose position in the ports string indicates the port number, and where <pin> is
either a base 10 number, a base 8 number preceded by "0", or a base 16 number
preceded by "0x" or "0X", defined by the strtol library function.
*/

static int parse_pin(const char *s, const char *ports, int *port, int *pin)
{
  int i = 0;

  if (!s || !(*s)) return 0;

  /* Parse prefix in character range. */

  while (*ports)
  {
    if (s[0] == *ports)
    {
      *port = i;

      /* Parse pin number. */

      *pin = (int) strtol(s+1, NULL, 0);
      return !errno;
    }
    ports++; i++;
  }

  return 0;
}

/* Arguments: <SPI clock pin> <SPI data pin> <SPI enable pin> */

int main(int argc, char *argv[])
{
  int clock_port, clock_pin, data_port, data_pin, enable_port, enable_pin;

  if (argc < 4)
    return 1;

  /* Interpret the pin details. */

  if (!parse_pin(argv[1], "ABCD", &clock_port, &clock_pin))
    return 1;
  if (!parse_pin(argv[2], "ABCD", &data_port, &data_pin))
    return 1;
  if (!parse_pin(argv[3], "ABCD", &enable_port, &enable_pin))
    return 1;

  /* Obtain access to peripheral memory. */

  if (setup_memory())
    return 1;

  /* Configure the clock pin. */

  Gpio_jz4740_chip gpio_port_clock(gpio_virt_base, clock_port);

  gpio_port_clock.setup(clock_pin, Hw::Gpio_chip::Output, 0);

  /* Configure the data pin. */

  Gpio_jz4740_chip gpio_port_data(gpio_virt_base, data_port);

  gpio_port_data.setup(data_pin, Hw::Gpio_chip::Output, 0);

  /* Configure the enable pin. */

  Gpio_jz4740_chip gpio_port_enable(gpio_virt_base, enable_port);

  gpio_port_enable.setup(enable_pin, Hw::Gpio_chip::Output, 0);

  /* Create an object for SPI communication omitting any specific frequency. */

  Spi_gpio spi(&gpio_port_clock, clock_pin, &gpio_port_data, data_pin,
               &gpio_port_enable, enable_pin);

  /* Initialise and register a new server object. */

  SPI_server obj(&spi);

  /* Bind and start the IPC server loop. */

  if (ipc_server_loop_for(SPI, &obj, "spi"))
    return 1;

  return 0;
}
